/*/../bin/true; p="/tmp/$0.elf"; v=1; g++ -O42 "$0" -o "$p" && chmod 755 "$p" && ("$p"; v=$?; rm "$p"); exit $v;*/

/* memfill.cpp - memory allocator behavior tester
 * 2018 ed <irc.rizon.net>, MIT licensed
 * https://ocv.me/dev/memfill.cpp
 * 
 * usage: ./memfill.cpp
 * 
 * allocating memory:
 *   n  s>  n  s>  n  s>  [...]
 * 
 * freeing memory:
 *   d>  d>  d>  d>  [...]
 */

#include <iostream>
#include <string>
#include <vector>
#include <stdlib.h>
#include <string.h>

// pid=''; while true; do [ $pid ] || pid=$(pidof memfill); [ $pid ] || { echo 'x'; sleep 1; continue; }; grep VmRSS /proc/$pid/status || pid='' ; sleep 0.2; done

int getint(std::string cmd, int min, int max)
{
	if (cmd.size() < 2)
	{
		std::cout << "command needs a numeric argument\n";
		return -1;
	}
	
	if (cmd[1] == '<')
		return min;
	
	if (cmd[1] == '>')
		return max;
	
	// ensure all characters after first is a digit
	std::string::const_iterator it = cmd.begin() + 1;
	while (it != cmd.end() && std::isdigit(*it)) ++it;
	if (cmd.empty() || it != cmd.end())
	{
		std::cout << "bad argument; must be a number OR '<' for min OR '>' for max: " << cmd.substr(1) << "\n";
		return -1;
	}
	int ret = atol(cmd.substr(1).c_str());
	if (ret < min)
	{
		std::cout << "argument below minimum: " << ret << " < " << min << "\n";
		return -1;
	}
	if (ret > max)
	{
		std::cout << "argument above maximum: " << ret << " > " << max << "\n";
		return -1;
	}
	return ret;
}

int main(int argc, char** argv)
{
	size_t blocksize = 1024;
	blocksize *= 1024;
	blocksize *= 16;
	std::cout << "blocksize = " << blocksize << "\n";
	
	uint64_t malloced = 0;

	std::vector<char*> blocks;
	while (true)
	{
		std::string cmd;
		std::cout << "\n[n]ew [r]ead [p]oke [s]et [u]nset [d]elete [q]uit\ncmd> ";
		if (!std::getline(std::cin, cmd))
			break;
		
		if (cmd.size() < 1)
			continue;
		
		if (cmd == "q")
			break;
		
		if (cmd[0] == 'n')
		{
			char* block = (char*)malloc(blocksize);
			blocks.push_back(block);
			malloced += blocksize;
			std::cout << "created block " << blocks.size() <<
				", " << (malloced/(1024*1024)) << "MB malloced\n";
			
			continue;
		}

		int idx = getint(cmd, 0, blocks.size() - 1);
		if (idx < 0)
			continue;

		if (cmd[0] == 'r')
		{
			std::cout << "reading values from block " << idx << "\n";
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
			{
				size_t v = *((int*)&(blocks[idx][ofs]));
				v += 1;
			}
			continue;
		}

		if (cmd[0] == 'p')
		{
			std::cout << "setting 256 values (1MB) in block " << idx << "\n";
			for (size_t ofs = 0; ofs < 0x100000; ofs += 0x1000)
				blocks[idx][ofs] = ofs/0x1000;

			continue;
		}

		if (cmd[0] == 's')
		{
			std::cout << "setting values in block " << idx << "\n";
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
				*((int*)&(blocks[idx][ofs])) = ofs;
			
			if (false)
				for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
					std::cout << std::hex << *((int*)&(blocks[idx][ofs])) << " ";

			continue;
		}

		if (cmd[0] == 'u')
		{
			std::cout << "blanking values in block " << idx << "\n";
			size_t nil = 0;
			for (size_t ofs = 0; ofs < blocksize; ofs += 1024)
				*((int*)&(blocks[idx][ofs])) = nil;

			continue;
		}

		if (cmd[0] == 'd')
		{
			malloced -= blocksize;
			std::cout << "erasing block at index " << idx <<
				", " << (malloced/(1024*1024)) << "MB malloced\n";
			
			char* block = blocks[idx];
			free(block);
			blocks.erase(blocks.begin() + idx);

			continue;
		}
	}
	
	while (!blocks.empty())
	{
		free(blocks.back());
		blocks.pop_back();
	}
}
